@@ -1,6 +1,6 @@
 /*
   Copyright (C) 2009-2015  Brazil
-  Copyright (C) 2018-2024  Sutou Kouhei <kou@clear-code.com>
+  Copyright (C) 2018-2025  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -19,6 +19,7 @@
 
 #include "grn_index_column.h"
 #include "grn_ii.h"
+#include "grn_float.h"
 #include "grn_hash.h"
 
 #include <string.h>
@@ -248,6 +249,74 @@ grn_index_column_rebuild(grn_ctx *ctx, grn_obj *index_column)
   GRN_API_RETURN(ctx->rc);
 }
 
+static const grn_log_level grn_index_column_diff_progress_log_level_default =
+  GRN_LOG_DEBUG;
+
+grn_rc
+grn_index_column_diff_options_init(grn_ctx *ctx,
+                                   grn_index_column_diff_options *options)
+{
+  options->progress_log_level =
+    grn_index_column_diff_progress_log_level_default;
+  return GRN_SUCCESS;
+}
+
+grn_rc
+grn_index_column_diff_options_fin(grn_ctx *ctx,
+                                  grn_index_column_diff_options *options)
+{
+  return GRN_SUCCESS;
+}
+
+grn_index_column_diff_options *
+grn_index_column_diff_options_open(grn_ctx *ctx)
+{
+  GRN_API_ENTER;
+  grn_index_column_diff_options *options =
+    GRN_MALLOC(sizeof(grn_index_column_diff_options));
+  if (!options) {
+    GRN_API_RETURN(NULL);
+  }
+  grn_rc rc = grn_index_column_diff_options_init(ctx, options);
+  if (rc != GRN_SUCCESS) {
+    GRN_FREE(options);
+    GRN_API_RETURN(NULL);
+  }
+  GRN_API_RETURN(options);
+}
+
+grn_rc
+grn_index_column_diff_options_close(grn_ctx *ctx,
+                                    grn_index_column_diff_options *options)
+{
+  GRN_API_ENTER;
+  grn_rc rc = GRN_SUCCESS;
+  if (options) {
+    rc = grn_index_column_diff_options_fin(ctx, options);
+    GRN_FREE(options);
+  }
+  GRN_API_RETURN(rc);
+}
+
+grn_log_level
+grn_index_column_diff_options_get_progress_log_level(
+  grn_ctx *ctx, grn_index_column_diff_options *options)
+{
+  if (options) {
+    return options->progress_log_level;
+  } else {
+    return grn_index_column_diff_progress_log_level_default;
+  }
+}
+
+grn_rc
+grn_index_column_diff_options_set_progress_log_level(
+  grn_ctx *ctx, grn_index_column_diff_options *options, grn_log_level level)
+{
+  options->progress_log_level = level;
+  return GRN_SUCCESS;
+}
+
 typedef enum {
   GRN_INDEX_COLUMN_DIFF_TYPE_REMAINS,
   GRN_INDEX_COLUMN_DIFF_TYPE_MISSINGS,
@@ -288,6 +357,7 @@ typedef struct {
 } grn_index_column_diff_posting_list;
 
 typedef struct {
+  grn_index_column_diff_options *options;
   grn_obj *lexicon;
   grn_ii *ii;
   struct {
@@ -326,8 +396,11 @@ typedef struct {
 } grn_index_column_diff_data;
 
 static void
-grn_index_column_diff_data_init(grn_ctx *ctx, grn_index_column_diff_data *data)
+grn_index_column_diff_data_init(grn_ctx *ctx,
+                                grn_index_column_diff_data *data,
+                                grn_index_column_diff_options *options)
 {
+  data->options = options;
   GRN_PTR_INIT(&(data->source_columns), GRN_OBJ_VECTOR, GRN_ID_NIL);
   GRN_VOID_INIT(&(data->buffers.value));
   GRN_UINT32_INIT(&(data->buffers.postings), GRN_OBJ_VECTOR);
@@ -372,7 +445,8 @@ grn_index_column_diff_init_progress(grn_ctx *ctx,
   data->progress.n_records = grn_table_size(ctx, data->source_table);
   data->progress.i = 0;
   data->progress.interval = 10000;
-  data->progress.log_level = GRN_LOG_DEBUG;
+  data->progress.log_level =
+    grn_index_column_diff_options_get_progress_log_level(ctx, data->options);
   double n_records_digits = ceil(log10(data->progress.n_records + 1));
   data->progress.n_records_digits = (int)n_records_digits;
   grn_timeval_now(ctx, &(data->progress.start_time));
@@ -443,7 +517,12 @@ grn_index_column_diff_progress(grn_ctx *ctx, grn_index_column_diff_data *data)
        current_time.tv_nsec / GRN_TIME_NSEC_PER_SEC_F) -
       ((double)(previous_time->tv_sec) +
        previous_time->tv_nsec / GRN_TIME_NSEC_PER_SEC_F);
-    const double throughput = interval / current_interval_seconds;
+    double throughput;
+    if (grn_float_is_zero(current_interval_seconds)) {
+      throughput = interval;
+    } else {
+      throughput = interval / current_interval_seconds;
+    }
     const double remained_seconds =
       elapsed_seconds + ((n_records - i) / throughput);
     const char *elapsed_unit = NULL;
@@ -583,7 +662,7 @@ grn_index_column_diff_posting_list_init(
 }
 
 static void
-grn_index_column_diff_posting_list_fin(
+grn_index_column_diff_posting_list_flush(
   grn_ctx *ctx,
   grn_index_column_diff_data *data,
   grn_index_column_diff_posting_list *posting_list)
@@ -597,7 +676,6 @@ grn_index_column_diff_posting_list_fin(
                    remains,
                    GRN_BULK_HEAD(look_ahead) + offset,
                    GRN_BULK_VSIZE(look_ahead) - offset);
-    GRN_OBJ_FIN(ctx, look_ahead);
   }
 
   grn_ii_cursor *cursor = posting_list->cursor;
@@ -607,8 +685,14 @@ grn_index_column_diff_posting_list_fin(
     if (with_position) {
       while (!posting_list->need_cursor_next ||
              grn_ii_cursor_next(ctx, cursor)) {
+        if (ctx->rc != GRN_SUCCESS) {
+          return;
+        }
         grn_posting *posting;
         while ((posting = grn_ii_cursor_next_pos(ctx, cursor))) {
+          if (ctx->rc != GRN_SUCCESS) {
+            return;
+          }
           posting_list->need_cursor_next = false;
           GRN_UINT32_PUT(ctx, remains, posting->rid);
           if (with_section) {
@@ -621,14 +705,15 @@ grn_index_column_diff_posting_list_fin(
     } else {
       grn_posting *posting;
       while ((posting = grn_ii_cursor_next(ctx, cursor))) {
+        if (ctx->rc != GRN_SUCCESS) {
+          return;
+        }
         GRN_UINT32_PUT(ctx, remains, posting->rid);
         if (with_section) {
           GRN_UINT32_PUT(ctx, remains, posting->sid);
         }
       }
     }
-    grn_ii_cursor_close(ctx, cursor);
-    posting_list->cursor = NULL;
   }
 
   if (GRN_BULK_VSIZE(remains) > 0 || GRN_BULK_VSIZE(missings) > 0) {
@@ -650,8 +735,24 @@ grn_index_column_diff_posting_list_fin(
       }
     }
   }
-  GRN_OBJ_FIN(ctx, remains);
-  GRN_OBJ_FIN(ctx, missings);
+}
+
+static void
+grn_index_column_diff_posting_list_fin(
+  grn_ctx *ctx,
+  grn_index_column_diff_data *data,
+  grn_index_column_diff_posting_list *posting_list)
+{
+  if (ctx->rc == GRN_SUCCESS) {
+    grn_index_column_diff_posting_list_flush(ctx, data, posting_list);
+  }
+  GRN_OBJ_FIN(ctx, &(posting_list->remains));
+  GRN_OBJ_FIN(ctx, &(posting_list->missings));
+  GRN_OBJ_FIN(ctx, &(posting_list->look_ahead));
+  if (posting_list->cursor) {
+    grn_ii_cursor_close(ctx, posting_list->cursor);
+    posting_list->cursor = NULL;
+  }
 }
 
 static void
@@ -909,8 +1010,14 @@ grn_index_column_diff_process_token_id(grn_ctx *ctx,
     if (with_position) {
       while (!posting_list->need_cursor_next ||
              grn_ii_cursor_next(ctx, cursor)) {
+        if (ctx->rc != GRN_SUCCESS) {
+          return;
+        }
         grn_posting *posting;
         while ((posting = grn_ii_cursor_next_pos(ctx, cursor))) {
+          if (ctx->rc != GRN_SUCCESS) {
+            return;
+          }
           posting_list->need_cursor_next = false;
           const grn_index_column_diff_compared compared =
             grn_index_column_diff_compare_posting(ctx, data, posting);
@@ -961,6 +1068,9 @@ grn_index_column_diff_process_token_id(grn_ctx *ctx,
     } else {
       grn_posting *posting;
       while ((posting = grn_ii_cursor_next(ctx, cursor))) {
+        if (ctx->rc != GRN_SUCCESS) {
+          return;
+        }
         const grn_index_column_diff_compared compared =
           grn_index_column_diff_compare_posting(ctx, data, posting);
         GRN_LOG(ctx,
@@ -1222,13 +1332,16 @@ grn_index_column_diff_compute(grn_ctx *ctx, grn_index_column_diff_data *data)
 }
 
 grn_rc
-grn_index_column_diff(grn_ctx *ctx, grn_obj *index_column, grn_obj **diff)
+grn_index_column_diff_full(grn_ctx *ctx,
+                           grn_obj *index_column,
+                           grn_index_column_diff_options *options,
+                           grn_obj **diff)
 {
   grn_index_column_diff_data data = {0};
 
   GRN_API_ENTER;
 
-  grn_index_column_diff_data_init(ctx, &data);
+  grn_index_column_diff_data_init(ctx, &data, options);
   if (ctx->rc != GRN_SUCCESS) {
     GRN_API_RETURN(ctx->rc);
   }
@@ -1379,6 +1492,12 @@ exit:
   GRN_API_RETURN(ctx->rc);
 }
 
+grn_rc
+grn_index_column_diff(grn_ctx *ctx, grn_obj *index_column, grn_obj **diff)
+{
+  return grn_index_column_diff_full(ctx, index_column, NULL, diff);
+}
+
 grn_obj *
 grn_index_column_get_tokenizer(grn_ctx *ctx, grn_obj *index_column)
 {
